package org.esupportail.cas.server.handlers.nis;

import java.util.Hashtable;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.directory.InitialDirContext;

import org.dom4j.Element;
import org.esupportail.cas.server.util.RedundantHandler;
import org.esupportail.cas.server.util.Server;
import org.esupportail.cas.server.util.crypt.Crypt;
import org.esupportail.cas.server.util.log.Log;

/**
 * This class implements a NIS (Network Information Service) server.
 *
 * @author Pascal Aubry <pascal.aubry at univ-rennes1.fr>
 */
public final class NisServer extends Server {
	
	/**
	 * The server hostname or IP address.
	 */
	private String host;

	/**
	 * Constructor.
	 *
	 * @param handlerDebug debugging mode of the handler
	 * @param handler      the handler the server will be used by
	 * @param serverElement the XML element that declares the server 
	 * @throws Exception Exception
	 */
	public NisServer(
			final Boolean handlerDebug,
			final RedundantHandler handler,
			final Element serverElement) throws Exception {
		super(handlerDebug, handler, serverElement);
		traceBegin();

		host = getServerSubElementContent(serverElement, "host", true/*needed*/);
		trace("host = " + host);

		traceEnd();
	}
	
	
	/**
	 * Try to authenticate a user (by searching into a NIS domain).
	 *
	 * @param username the user's name
	 * @param password the user's password
	 *
	 * @return Server.AUTHENTICATE_SUCCESS, Server.AUTHENTICATE_NOAUTH
	 * or Server.AUTHENTICATE_FAILURE.
	 */	
	public int authenticate(final String username,
			final String password) {
		traceBegin();
		
		NisHandler handler = (NisHandler) getHandler();
		String url = "nis://" + host + "/" + handler.getDomain();
		String map = handler.getMap();
		
		try {
			trace("Connecting to the NIS domain...");
			Hashtable hashtable = new Hashtable(5, 0.75f);
			hashtable.put(Context.INITIAL_CONTEXT_FACTORY, "com.sun.jndi.nis.NISCtxFactory");
			hashtable.put(Context.PROVIDER_URL, url);
			hashtable.put(Context.SECURITY_AUTHENTICATION, "simple");
			InitialContext context = new InitialDirContext(hashtable);
			
			trace("Retrieving the information corresponding to the user...");
			String nisEntry = context.lookup("system/" + map + "/" + username).toString();
			
			// we've got all needed information, close the context
			context.close();
			
			trace("Username found, checking password (" + handler.getEncryption() + ")...");
			// extracting the encrypted password
			String[] nisFields = nisEntry.split(":");
			String nisEncryptedPassword = nisFields[1];
			
			// compare the passwords
			boolean match = Crypt.match(handler.getEncryption(), password, nisEncryptedPassword);
			
			if (Crypt.match(handler.getEncryption(), password, nisEncryptedPassword)) {
				trace("Password matches.");
				traceEnd("AUTHENTICATE_SUCCESS");
				return AUTHENTICATE_SUCCESS;
			} else {
				trace("Password does not match.");
				traceEnd("AUTHENTICATE_NOAUTH");
				return AUTHENTICATE_NOAUTH;
			}
		} catch (javax.naming.NoInitialContextException e) {
			Log.warn(e.toString());
			Log.warn("JNDI nis provider (nis.jar) is probably not installed");
			traceEnd("AUTHENTICATE_FAILURE");
			return AUTHENTICATE_FAILURE;
		} catch (javax.naming.ConfigurationException e) {
			Log.warn("Bad NIS configuration: " + e.getMessage());
			traceEnd("AUTHENTICATE_FAILURE");
			return AUTHENTICATE_FAILURE;
		} catch (javax.naming.CommunicationException e) {
			Log.warn("NIS server not responding.");
			traceEnd("AUTHENTICATE_FAILURE");
			return AUTHENTICATE_FAILURE;
		} catch (javax.naming.CannotProceedException e) {
			Log.warn("Can not proceed: " + e.getMessage());
			traceEnd("AUTHENTICATE_NOAUTH");
			return AUTHENTICATE_NOAUTH;
		} catch (javax.naming.NameNotFoundException e) {
			trace("Username not found: " + e.getMessage());
			traceEnd("AUTHENTICATE_NOAUTH");
			return AUTHENTICATE_NOAUTH;
		} catch (Exception e) {
			Log.warn("Failure: " + e.toString());
			traceEnd("AUTHENTICATE_FAILURE");
			return AUTHENTICATE_FAILURE;
		}
	}
	
}

