package org.esupportail.cas.server.util;

import org.dom4j.Element;
import org.esupportail.cas.server.util.log.Debug;

/**
 * This abstract class implements a generic server class, inherited 
 * by DatabaseServer and LDAPServer.
 *
 * @author Pascal Aubry <pascal.aubry at univ-rennes1.fr>
 */

public abstract class Server extends Debug {
	
	/**
	 * The error set by the connect() method on success.
	 */
	protected static final int CONNECT_SUCCESS = 0;
	/**
	 * The error code set by the connect() method when the connection 
	 * with the server could not be achieved because of the bind
	 * information provided (the server responded but refused the 
	 * connection). Returned by getConnectError().
	 */
	protected static final int CONNECT_NOAUTH  = 1;
	/**
	 * The error code set by the connect() method when the connection 
	 * with the server could not be achieved because of a service failure.
	 * Returned by getConnectError().
	 */
	protected static final int CONNECT_FAILURE = 2;
	/**
	 * an integer to store the error code of the connect() calls, set by
	 * setConnectError(), retrieved by getConnectError().
	 */
	private int connectError = CONNECT_SUCCESS;
	/**
	 * The result of the authenticate() method on success (when the 
	 * user has been authenticated).
	 */
	protected static final int AUTHENTICATE_SUCCESS = 0;
	/**
	 * The result of the authenticate() method when the user could not 
	 * be authenticated (but the server was accessed without any problem).
	 * When testing many servers, the next servers will not be called.
	 */
	protected static final int AUTHENTICATE_NOAUTH = 1;
	/**
	 * The result of the authenticate() method on failure (when the server 
	 * could not be accessed for instance). When testing many servers, the
	 * next server will be called.
	 */
	protected static final int AUTHENTICATE_FAILURE = 2;

	/**
	 * The handler the server will be used for.
	 */
	private RedundantHandler handler;

	
	/**
	 * Constructor.
	 *
	 * @param handlerDebug debugging mode
	 * @param parentHandler the handler the server will be used by
	 * @param serverElement an XML element corresponding to the server configuration
	 */
	public Server(
			final Boolean handlerDebug,
			final RedundantHandler parentHandler,
			final Element serverElement) {
		super(elementDebugValue(serverElement, handlerDebug.booleanValue()));
		handler = parentHandler;
	}
	
	/**
	 * Get the handler the server will be used by.
	 *
	 * @return a RedundantHandler object.
	 */
	protected final RedundantHandler getHandler() {
		return handler;
	}
	
	/**
	 * Try to authenticate a user.
	 *
	 * @param username the user's name
	 * @param password the user's password
	 *
	 * @return Server.AUTHENTICATE_SUCCESS, Server.AUTHENTICATE_NOAUTH
	 * or Server.AUTHENTICATE_FAILURE.
	 */	
	protected abstract int authenticate(final String username,
			final String password);
	
	/**
	 * Set the error code of the last connect() call.
	 *
	 * @param errCode an integer equal to Server.CONNECT_SUCCESS, 
	 * Server.CONNECT_NOAUTH or Server.CONNECT_FAILURE.
	 */
	protected final void setConnectError(final int errCode) {
		connectError = errCode;
	}
	
	/**
	 * Retrieve the error code of the last connect() call.
	 *
	 * @return Server.CONNECT_SUCCESS, Server.CONNECT_NOAUTH
	 * or Server.CONNECT_FAILURE.
	 */
	protected final int getConnectError() {
		int error = connectError;
		connectError = CONNECT_SUCCESS;
		return error;
	}
	
	/**
	 * Retrieve the content of a server sub-element.
	 * 
	 * @param serverElement the XML element representing the server in the configuration
	 * @param elementName the name of the XML element to look for
	 * @param needed false if the element can be absent or empty, true otherwise
	 * @return a String
	 * @throws Exception Exception
	 */
	protected final String getServerSubElementContent(
			final Element serverElement, 
			final String elementName, 
			final boolean needed) throws Exception {
		Element element = serverElement.element(elementName);
		String str;
		if (element == null) {
			str = "";
		} else {
			str = element.getTextTrim();
		}
		if (needed && str.equals("")) {
			traceThrow(new MisconfiguredHandlerException(
					"A non empty nested \"" + elementName + "\" element is needed to configure \"" 
					+ getClass().getName() 
					+ "\" servers."
					));
		}
		return str;
	}


}
